#define HAVE_CLOCK
#define HAVE_64BIT_INTEGER
#define BLOCK_SIZE      8
#define NUM_CLASS       -1
#define NUM_GROUP       16
#define PRD_ORDER       -1
#define COEF_PRECISION  6
#define MAX_UPARA       512
#define NUM_UPELS       6
#define UPEL_DIST       2
#define MAX_ITERATION   100
#define EXTRA_ITERATION 10
#define NUM_PMODEL      1
#define MAX_TOTFREQ     (1 << 20)	/* must be less than RANGE_BOT */
#define MIN_FREQ        1
#define VLC_MAXLEN      32
#define MAGIC_NUMBER    ('M' << 8) + 'R'
#define BANNER          "encmrp/decmrp version %.1f (Oct. 2002)"
#define VERSION         3
#define uchar           unsigned char
#define uint            unsigned int
#define img_t           unsigned char
#define cost_t          float
#ifdef HAVE_64BIT_INTEGER
#  define RANGE_SIZE 64
#  ifdef __WIN32__
#    define range_t unsigned __int64
#  else
#    define range_t unsigned long long
#  endif
#else
#  define RANGE_SIZE 32
#  define range_t unsigned int
#endif
#define RANGE_TOP  ((range_t)1 << (RANGE_SIZE - 8))
#define RANGE_BOT  ((range_t)1 << (RANGE_SIZE - 16))

typedef struct {
    int height;
    int width;
    int maxval;
    img_t **val;
} IMAGE;

typedef struct {
    int size;
    uint totfreq;
    uint *freq;
    uint *cumfreq;
    double log_tf;
} PMODEL;

typedef struct {
    int size;
    int max_len;
    int *len;
    int *index;
    int *off;
    uint *code;
} VLC;

typedef struct {
    range_t low;
    range_t code;
    range_t range;
    int     maxsym;
} RANGECODER;

typedef struct {
    int y, x;
} POINT;

typedef struct {
    int height;
    int width;
    int maxval;
    int num_class;
    int num_group;
    int prd_order;
    int num_pmodel;
    int coef_precision;
    int f_huffman;
    int **predictor;
    int **th;
    cost_t **cost;
    int **upara;
    img_t ***ref_pels;
    img_t **abs_e;
    img_t **idx_E;
    img_t **cost_e;
    img_t **econv;
    uchar **class;
    uchar **group;
    int *pm_idx;
    PMODEL **pmodels;
    VLC **vlcs;
    RANGECODER *rc;
    double *sigma;
} ENCODER;

typedef struct {
    int version;
    int height;
    int width;
    int maxval;
    int num_comp;
    int num_class;
    int num_group;
    int prd_order;
    int num_pmodel;
    int coef_precision;
    int f_huffman;
    int **predictor;
    int **th;
    img_t **idx_E;
    uchar **class;
    int *pm_idx;
    PMODEL **pmodels;
    VLC **vlcs;
    RANGECODER *rc;
    double *sigma;
} DECODER;

/* common.c */
FILE *fileopen(char *, char *);
void *alloc_mem(size_t);
void **alloc_2d_array(int, int, int);
IMAGE *alloc_image(int, int, int);
int *gen_hufflen(uint *, int, int);
void gen_huffcode(VLC *);
VLC *make_vlc(uint *, int, int);
VLC ***init_vlc(int, int, int *, int);
void free_vlc(VLC *);
VLC **init_vlcs(PMODEL **, int, int);
PMODEL **init_pmodels(int, int, int *, double *, int);
int calc_u(img_t **, int, int, int);
int e2E(int, int, int, int);
int E2e(int, int, int, int);
double cpu_time(void);

/* rc.c */
RANGECODER *rc_init(int);
void rc_encode(FILE *, RANGECODER *, uint, uint, uint);
void rc_finishenc(FILE *, RANGECODER *);
int rc_decode(FILE *, RANGECODER *, uint *, uint *, uint);
void rc_startdec(FILE *, RANGECODER *);
